//==========================================================================
//
//      plf_misc.c
//
//      HAL platform miscellaneous functions
//
//==========================================================================
//####ECOSGPLCOPYRIGHTBEGIN####
// -------------------------------------------
// This file is part of eCos, the Embedded Configurable Operating System.
// Copyright (C) 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
//
// eCos is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later version.
//
// eCos is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
// for more details.
//
// You should have received a copy of the GNU General Public License along
// with eCos; if not, write to the Free Software Foundation, Inc.,
// 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
//
// As a special exception, if other files instantiate templates or use macros
// or inline functions from this file, or you compile this file and link it
// with other works to produce a work based on this file, this file does not
// by itself cause the resulting work to be covered by the GNU General Public
// License. However the source code for this file must still be made available
// in accordance with section (3) of the GNU General Public License.
//
// This exception does not invalidate any other reasons why a work based on
// this file might be covered by the GNU General Public License.
//
// Alternative licenses for eCos may be arranged by contacting Red Hat, Inc.
// at http://sources.redhat.com/ecos/ecos-license/
// -------------------------------------------
//####ECOSGPLCOPYRIGHTEND####
//==========================================================================
//#####DESCRIPTIONBEGIN####
//
// Author(s):    msieweke
// Contributors: nickg, jlarmour, dmoseley, jskov
// Date:         2003-06-16
// Purpose:      HAL miscellaneous functions
// Description:  This file contains miscellaneous functions provided by the
//               HAL.
//
//####DESCRIPTIONEND####
//
//========================================================================*/

#define CYGARC_HAL_COMMON_EXPORT_CPU_MACROS
#include <pkgconf/hal.h>

#include <cyg/infra/cyg_type.h>         // Base types
#include <cyg/infra/cyg_trac.h>         // tracing macros
#include <cyg/infra/cyg_ass.h>          // assertion macros
#include <cyg/infra/diag.h>             // diag output

#include <cyg/hal/hal_arch.h>           // architectural definitions
#include <cyg/hal/hal_intr.h>           // Interrupt handling
#include <cyg/hal/hal_cache.h>          // Cache handling
#include <cyg/hal/hal_if.h>
#include <cyg/hal/hal_diag.h>
#include <cyg/hal/bcm33xx_regs.h>

#include <cyg/kernel/kapi.h>


//--------------------------------------------------------------------------

#define MAP_VECTOR_TO_BRCM_EXT_INTERRUPT( vector ) \
                  ((vector) - CYGNUM_HAL_INTERRUPT_PERIPH - 1)

#define MAP_BRCM_EXT_INTERRUPT_TO_VECTOR( int ) \
                  ((int) + CYGNUM_HAL_INTERRUPT_PERIPH + 1)

//--------------------------------------------------------------------------

externC void diag_write_string( const char* );
static void hal_init_periph_interrupt( void );
static void hal_init_external_interrupts( void );
static void hal_init_software_interrupts( void );

//--------------------------------------------------------------------------

// We want to support multiple clocks on different platforms, so we can't use
// a hard-coded constant.  This global is defined with weak linkage so it can
// be overridden in the application.
cyg_uint32 hal_mips_rtc_period __attribute__ ((weak)) = 
    (CYGHWR_HAL_MIPS_BCM33XX_CPU_CLOCK / 2) / CYGNUM_HAL_RTC_DENOMINATOR;

// The eCos library is built for generic 33xx devices.  The 3368 uses different
// addresses for cores and registers, so we make these variables with weak
// linkage to be substituted in the application.
volatile IntControl *INTC  __attribute__ ((weak)) = (IntControl *)  BCM_INTC_BASE;
volatile Uart       *UART  __attribute__ ((weak)) = (Uart *)        BCM_UART0_BASE;
volatile Uart       *UART1 __attribute__ ((weak)) = (Uart *)        BCM_UART1_BASE;
volatile Sdram      *SDRAM __attribute__ ((weak)) = (Sdram*)        BCM_SDRAM_BASE;
volatile Timer      *TIMER __attribute__ ((weak)) = (Timer*)        BCM_TIMER_BASE;

//--------------------------------------------------------------------------

//------------------------------------------------------------------------
// This function is specifically intended to be overridden in the BSP for
// any chip which requires special init code.
void chip_specific_init(void) __attribute__ ((weak));
void chip_specific_init(void)
{
}


void hal_platform_init(void)
{
    // Set up eCos/ROM interfaces
    hal_if_init();

    // Do any chip-specific initialization - in the app-level code.
    chip_specific_init();

#if defined(CYGPKG_CYGMON)
    diag_write_string(" CYGMON");
#elif defined(CYGPKG_REDBOOT)
    diag_write_string(" RedBoot");
#else
    diag_write_string(" eCos");
#endif

    hal_init_periph_interrupt();
    hal_init_external_interrupts();
    hal_init_software_interrupts();
}


//------------------------------------------------------------------------
// This function is specifically intended to be overridden in the BSP for
// any chip which requires special init code.
void chip_specific_cache_init(void) __attribute__ ((weak));
void chip_specific_cache_init(void)
{
}


//------------------------------------------------------------------------
// There are 32 internal interrupt sources routed through the periphal interrupt,
// so we want to be have an ISR for each one (let's call them sub-ISRs).  Each can
// specify a data value to be passed in as a parameter.  For example a single 
// sub-ISR could handle two or more sources, and the data would indicate which 
// source caused the interrupt.

static void   (*subIsr[32])( void *data )               = { NULL };
static void    *subIsrData[32]                          = { NULL };

//------------------------------------------------------------------------
// Initial interrupt service routine for peripheral interrupt. This interrupt is
// the sink for all all internal interrupts - most of which should be disabled
// when this ISR is attached.

static cyg_handle_t   hal_isr_periph_handle;
static cyg_interrupt  hal_isr_periph_interrupt;

static cyg_uint32 hal_isr_periph( cyg_vector_t    vector,
                             cyg_addrword_t  data)
{
    // Disable the interrupt until after the DSR runs.
    HAL_INTERRUPT_MASK( vector );

    // This ISR's only function is to kick off the DSR.
    return( CYG_ISR_HANDLED | CYG_ISR_CALL_DSR );
}

static void hal_dsr_periph( cyg_vector_t    vector,
                       cyg_ucount32    count,
                       cyg_addrword_t  data )
{
    cyg_uint32  activeInterrupts;
    cyg_uint32  currentInterrupt;
    int         i;

#if 0
    // When we initially brought up this HAL, we were getting spurious interrupts
    // for ISR 0 with no active peripheral interrupts.  This is a test to see
    // if that's still happening.
    static int  spuriousInterruptCount = 0;

    if (INTC->IrqStatus & INTC->IrqEnableMask == 0)
    {
        spuriousInterruptCount += 1;
        if (spuriousInterruptCount > 10)
        {
            spuriousInterruptCount = 0;
            diag_write_char('@');
        }
    }
#endif

    // Mask the asserted interrupts with the enable mask to find which are active.
    while (( activeInterrupts = INTC->IrqStatus & INTC->IrqEnableMask ) != 0 )
    {
#if 1
        // We start looking at the leftmost interrupt bit - bit 31.  This is
        // optimized by using assembler to skip leading zeroes:
        int  skipCount;
        asm( "clz %0, %1" : "=r" (skipCount) : "r" (activeInterrupts) );
        currentInterrupt = 1 << (31 - skipCount);

        for ( i = (31 - skipCount); i >= 0; i-- )
#else
        currentInterrupt = 1 << 31;
    
        // Test all interrupt bits starting with bit 31.  If an interrupt is active
        // and there is an interrupt handler registered, call the handler.
        for ( i = 31; i >= 0; i-- )
#endif
        {
            if (( activeInterrupts & currentInterrupt ) != 0)
            {
                // Disable the interrupt.  The handler should re-enable it.
                INTC->IrqEnableMask &= ~currentInterrupt;
                if (subIsr[i] != NULL)
                {
                    subIsr[i]( subIsrData[i] );
                }
            }
            currentInterrupt >>= 1;
        }
    }

    // The manual says this has no effect.  It was added to try to eliminate the
    // spurious interrupts we're seeing.
//    HAL_INTERRUPT_ACKNOWLEDGE( vector );

    // Re-enable our interrupt, which was disabled in the ISR.
    HAL_INTERRUPT_UNMASK( vector );
}

//------------------------------------------------------------------------
// Configure our ISR for the peripheral interrupt. The application may replace
// this with its own ISR at a later time.

static void hal_init_periph_interrupt()
{
    cyg_interrupt_create( CYGNUM_HAL_INTERRUPT_PERIPH,
                          0,                    // priority (unused in MIPS)
                          0,                    // Data item passed to ISR
                          &hal_isr_periph,
                          &hal_dsr_periph,
                          &hal_isr_periph_handle,
                          &hal_isr_periph_interrupt );
    cyg_interrupt_attach( hal_isr_periph_handle );
    HAL_INTERRUPT_UNMASK( CYGNUM_HAL_INTERRUPT_PERIPH );
}

//------------------------------------------------------------------------
// Register a sub-interrupt handler for the peripheral ISR. There are 32 possible
// handlers, although some are not used on some CPUs.

bool hal_map_periph_interrupt( cyg_uint32  subIsrNum,
                               void      (*newSubIsr)(void *), 
                               void       *newData )
{
    if (subIsrNum < 32)
    {
        subIsr[subIsrNum]     = newSubIsr;
        subIsrData[subIsrNum] = newData;
        return true;
    }
    else
    {
        return false;
    }
}

//------------------------------------------------------------------------
// Enable a sub-interrupt under the peripheral ISR. This is a simple 32-bit mask,
// and the interrupt numbers are defined as the bit offset.

void hal_enable_periph_interrupt( cyg_uint32 subIsrNum )
{
    if (subIsrNum < 32)
    {
        INTC->IrqEnableMask |= (1 << subIsrNum);
    }
}

//------------------------------------------------------------------------
// Disable a sub-interrupt under the peripheral ISR. This is a simple 32-bit mask,
// and the interrupt numbers are defined as the bit offset.

void hal_disable_periph_interrupt( cyg_uint32 subIsrNum )
{
    if (subIsrNum < 32)
    {
        INTC->IrqEnableMask &= ~(1 << subIsrNum);
    }
}

//------------------------------------------------------------------------
// There are four external interrupts (0-3) on BCM33xx chips, which map as 
// vectors 1-4.  We allow the app code to map a simple function call to
// each vector, and we maintain the eCos interrupt data structures.

static void      (*extIsr[4])( void *data ) = { NULL };
static void       *extIsrData[4]            = { NULL };

//------------------------------------------------------------------------
static cyg_uint32 hal_isr_ext( cyg_vector_t    vector, 
                               cyg_addrword_t  data)
{
//    hal_diag_write_char( 'x' );
    // Disable the interrupt until after the DSR runs.
    HAL_INTERRUPT_MASK( vector );

    // This ISR's only function is to kick off the DSR.
    return( CYG_ISR_HANDLED | CYG_ISR_CALL_DSR );
}


// This function is intended to be called from the eCos kernel ISR.  The
// external interrupts are mapped differently on the 3368, so we need to
// be able to override the default action.
void BcmHalClearExtInterrupt( int vector ) __attribute__ ((weak));
void BcmHalClearExtInterrupt( int vector )
{
    // Note that our external interrupts 0-3 are equivalent to eCos vectors
    // 3-6.  So we have to map "vector" to get a Broadcom interrupt number.
    INTC->extIrqMskandDetClr |= 1 << (MAP_VECTOR_TO_BRCM_EXT_INTERRUPT( vector ));
}


//------------------------------------------------------------------------
static void hal_dsr_ext( cyg_vector_t    vector, 
                         cyg_ucount32    count,
                         cyg_addrword_t  data )
{
    // Note that our external interrupts 0-3 are equivalent to eCos vectors
    // 3-6.  So we have to map "vector" to get a Broadcom interrupt number.
    int ExtIntNum = MAP_VECTOR_TO_BRCM_EXT_INTERRUPT( vector );

    // Acknowledge the interrupt, because the external "ISR" may unmask it.
    HAL_INTERRUPT_ACKNOWLEDGE( vector );

    // Clear the external interrupt, so we won't get another interrupt before
    // we're ready.
    BcmHalClearExtInterrupt( vector );

    // Call the associated external interrupt routine.
    if (extIsr[ExtIntNum] != NULL)
    {
        extIsr[ExtIntNum]( extIsrData[ExtIntNum] );
    }

    // The external ISR should re-enable the interrupt vector.
    // HAL_INTERRUPT_UNMASK( vector );
}


static cyg_handle_t   hal_isr_ext_1_handle;
static cyg_interrupt  hal_isr_ext_1_interrupt;

static cyg_handle_t   hal_isr_ext_2_handle;
static cyg_interrupt  hal_isr_ext_2_interrupt;

static cyg_handle_t   hal_isr_ext_3_handle;
static cyg_interrupt  hal_isr_ext_3_interrupt;

static cyg_handle_t   hal_isr_ext_4_handle;
static cyg_interrupt  hal_isr_ext_4_interrupt;

//------------------------------------------------------------------------
// Configure our ISRs for external interrupts.  The application may replace this with
// its own ISR at a later time.

static void hal_init_external_interrupts()
{
    cyg_interrupt_create( CYGNUM_HAL_INTERRUPT_3,  // vector 3 (external 0)
                          0,                    // priority (unused in MIPS)
                          0,                    // Data item passed to ISR
                          &hal_isr_ext,
                          &hal_dsr_ext,
                          &hal_isr_ext_1_handle,
                          &hal_isr_ext_1_interrupt );
    cyg_interrupt_attach( hal_isr_ext_1_handle );
    HAL_INTERRUPT_UNMASK( CYGNUM_HAL_INTERRUPT_3 );

    cyg_interrupt_create( CYGNUM_HAL_INTERRUPT_4,  // vector 4 (external 1)
                          0,                    // priority (unused in MIPS)
                          0,                    // Data item passed to ISR
                          &hal_isr_ext,
                          &hal_dsr_ext,
                          &hal_isr_ext_2_handle,
                          &hal_isr_ext_2_interrupt );
    cyg_interrupt_attach( hal_isr_ext_2_handle );
    HAL_INTERRUPT_UNMASK( CYGNUM_HAL_INTERRUPT_4 );

    cyg_interrupt_create( CYGNUM_HAL_INTERRUPT_5,  // vector 5 (external 2)
                          0,                    // priority (unused in MIPS)
                          0,                    // Data item passed to ISR
                          &hal_isr_ext,
                          &hal_dsr_ext,
                          &hal_isr_ext_3_handle,
                          &hal_isr_ext_3_interrupt );
    cyg_interrupt_attach( hal_isr_ext_3_handle );
    HAL_INTERRUPT_UNMASK( CYGNUM_HAL_INTERRUPT_5 );

    cyg_interrupt_create( CYGNUM_HAL_INTERRUPT_6,  // vector 6 (external 3)
                          0,                    // priority (unused in MIPS)
                          0,                    // Data item passed to ISR
                          &hal_isr_ext,
                          &hal_dsr_ext,
                          &hal_isr_ext_4_handle,
                          &hal_isr_ext_4_interrupt );
    cyg_interrupt_attach( hal_isr_ext_4_handle );
    HAL_INTERRUPT_UNMASK( CYGNUM_HAL_INTERRUPT_6 );
}

//------------------------------------------------------------------------
// Register an external interrupt handler.  There are 4 possible external
// sources, numbered 0-3.

bool hal_map_ext_interrupt( cyg_uint32  extIsrNum, 
                            void      (*newExtIsr)(void *), 
                            void       *newData )
{
    if (extIsrNum < 4)
    {
        extIsr[extIsrNum]     = newExtIsr;
        extIsrData[extIsrNum] = newData;
        return true;
    }
    else
    {
        return false;
    }
}

//------------------------------------------------------------------------
// Enable an external interrupt.  The vector is a number from 1 to 4

void hal_enable_ext_interrupt( cyg_uint32 extIsrNum )
{
    cyg_uint32 vector = MAP_BRCM_EXT_INTERRUPT_TO_VECTOR( extIsrNum );
    HAL_INTERRUPT_UNMASK( vector );
}

//------------------------------------------------------------------------
// Disable an external interrupt.  The vector is a number from 1 to 4

void hal_disable_ext_interrupt( cyg_uint32 extIsrNum )
{
    cyg_uint32 vector = MAP_BRCM_EXT_INTERRUPT_TO_VECTOR( extIsrNum );
    HAL_INTERRUPT_MASK( vector );
}

//------------------------------------------------------------------------
// There are two software interrupts (0-1) on BCM33xx chips.
// We allow the app code to map a simple function call to
// each vector, and we maintain the eCos interrupt data structures.

static void      (*swIsr[4])( void *data ) = { NULL };
static void       *swIsrData[4]            = { NULL };

//------------------------------------------------------------------------
static cyg_uint32 hal_isr_software( cyg_vector_t    vector,
                                    cyg_addrword_t  data)
{
//    hal_diag_write_char( 'x' );
    // Disable the interrupt until after the DSR runs.
    HAL_INTERRUPT_MASK( vector );

    // This ISR's only function is to kick off the DSR.
    return( CYG_ISR_HANDLED | CYG_ISR_CALL_DSR );
}




//------------------------------------------------------------------------
static void hal_dsr_software( cyg_vector_t    vector,
                              cyg_ucount32    count,
                              cyg_addrword_t  data )
{
    // Acknowledge the interrupt, because the sofware "ISR" may unmask it.
    HAL_INTERRUPT_ACKNOWLEDGE( vector );


    // Call the associated software interrupt routine.
    if (swIsr[vector] != NULL)
    {
        swIsr[vector]( swIsrData[vector] );
    }

}


static cyg_handle_t   hal_isr_sw_0_handle;
static cyg_interrupt  hal_isr_sw_0_interrupt;

static cyg_handle_t   hal_isr_sw_1_handle;
static cyg_interrupt  hal_isr_sw_1_interrupt;


//------------------------------------------------------------------------
// Configure our ISRs for software interrupts.  The application may replace this with
// its own ISR at a later time.

static void hal_init_software_interrupts()
{
    cyg_interrupt_create( CYGNUM_HAL_INTERRUPT_SOFTWARE_0,  // vector 0
                          0,                                // priority (unused in MIPS)
                          0,                                // Data item passed to ISR
                          &hal_isr_software,
                          &hal_dsr_software,
                          &hal_isr_sw_0_handle,
                          &hal_isr_sw_0_interrupt );
    cyg_interrupt_attach( hal_isr_sw_0_handle );
    HAL_INTERRUPT_UNMASK( CYGNUM_HAL_INTERRUPT_SOFTWARE_0 );

    cyg_interrupt_create( CYGNUM_HAL_INTERRUPT_SOFTWARE_1,  // vector 1
                          0,                                // priority (unused in MIPS)
                          0,                                // Data item passed to ISR
                          &hal_isr_software,
                          &hal_dsr_software,
                          &hal_isr_sw_1_handle,
                          &hal_isr_sw_1_interrupt );
    cyg_interrupt_attach( hal_isr_sw_1_handle );
    HAL_INTERRUPT_UNMASK( CYGNUM_HAL_INTERRUPT_SOFTWARE_1 );

}

//------------------------------------------------------------------------
// Register an software interrupt handler.  There are 2 possible software
// sources, numbered 0-1.

bool hal_map_software_interrupt( cyg_uint32  swIsrNum,
                                 void      (*newSwIsr)(void *),
                                 void       *newData )
{
    if (swIsrNum < 2)
    {
        swIsr[swIsrNum]     = newSwIsr;
        swIsrData[swIsrNum] = newData;
        return true;
    }
    else
    {
        return false;
    }
}


/*------------------------------------------------------------------------*/
/* Reset support                                                          */

void hal_bcm33xx_reset(void)
{
    diag_write_string( "(hal_bcm33xx_reset) Resetting...\n" );
//***** unfinished
    for(;;);                            // wait for it
}



//*****************************************************************************
// Calculate RAM size by writing test values to RAM.  This counts on the address
// being partially decoded, so RAM will be duplicated (RAM space)/(true size) times.
//
// DANGER: Since this function modifies memory owned by other tasks, it is
// not interrupt safe.  To guard against memory corruption, interrupts are
// disabled while it runs.
//
// We shouldn't have to worry about overwriting the stack and local variables,
// because the initial stack grows downward from location 0x10000.

cyg_uint32 CalculateRamSize(void) __attribute__ ((weak));
cyg_uint32 CalculateRamSize(void)
{
    unsigned long __state;
    cyg_uint16    RevID;

    RevID = INTC->RevID >> 16;
    // Assume the 3368 has exactly 16M of RAM.
    if ( RevID == 0x3368 || RevID == 0x0000 )
    {
        return 16 * 1024 * 1024;
    }

    HAL_DISABLE_INTERRUPTS(__state);

    // These are the base-2 logs of address space size (in megabytes) for values
    // in the memory base register/size field.
    static cyg_uint8  MemSizeValues[8] = { 1, 3, 4, 5, 6, 7, 8, 9 };

    cyg_uint32   MemSizeReg;          // Size of the memory address space - an enum
                                  // where 0=2M, 1=8M, 2=16M, 3=32M, 4=64M...
    cyg_uint32   DetectedRamSize = 0; // Set to 0 to eliminate compile warning.
    cyg_uint32   SavedValues[8];      // Saved values for 1M - 128M

    int                   i;
    int                   ProbeLimit;
    volatile cyg_uint32  *ProbeAddress;
    cyg_uint32            RamBase        = 0xa0000000; // RAM address 0 in non-cached space
    cyg_uint32            RamBaseValue;

    // The memory-base-register/memory-space says how large RAM _could_ be.  We 
    // can't probe outside this address space without getting a bus error.
    MemSizeReg  = SDRAM->memoryBase & SD_MEMSIZE_MASK;

    // Convert from the register value to the base-2 logarithm of the memory 
    // address space size in megabytes.
    ProbeLimit   = MemSizeValues[MemSizeReg];

    // Save the value at RAM location 0.
    ProbeAddress  = (cyg_uint32*) RamBase;
    RamBaseValue  = *ProbeAddress;

    // Save values at address 1M, 2M, 4M, 8M, etc.  Then store 0 in each location.
    // ( Remember that 1<<20 is 1M. )
    for (i = 0; i < ProbeLimit; i++)
    {
        ProbeAddress   = (cyg_uint32*) (RamBase + (1 << (i + 20)));
        SavedValues[i] = *ProbeAddress;
        *ProbeAddress  = 0;
    }

    #define PROBE_VALUE 0xf0f0f0f0

    // Write a non-zero value to RAM location 0.  Where RAM wraps around, this
    // value will show up above the top of memory.
    ProbeAddress  = (cyg_uint32*) RamBase;
    *ProbeAddress = PROBE_VALUE;
    
    // Start probing at memory address 1Meg.  Technically it's not necessary to
    // probe addresses at 1M or 4M, but this way is simpler.
    for (i = 0; i < ProbeLimit; i++)
    {
        ProbeAddress  = (cyg_uint32*) (RamBase + (1 << (i + 20)));
        if (*ProbeAddress == PROBE_VALUE)
        {
            break;
        }
        // Restore the saved value for this address.
        *ProbeAddress = SavedValues[i];
    }
    // RAM size is 2^i megabytes...
    DetectedRamSize = 1 << (i + 20);

    // Restore the value we found at RAM address 0.
    ProbeAddress  = (cyg_uint32*) RamBase;
    *ProbeAddress =  RamBaseValue;

    HAL_RESTORE_INTERRUPTS(__state);

    return DetectedRamSize;
}


//------------------------------------------------------------------------
// Run-time memory size info.

// Allow the application to set an arbitrary limit on memory usage.  If the
// application defines this variable, that's used as the size limit.
cyg_uint32 hal_mips_mem_limit __attribute__ ((weak)) = 64*1024*1024;

#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define MIN(a,b) ((a) < (b) ? (a) : (b))

cyg_uint8 *hal_mips_mem_real_region_top( cyg_uint8 *_regionend_ )
{
  
    static cyg_uint8  *memTop   = NULL;
    // Don't assume which segment to put the "memory top" value.  Use the
    // same segment as the one passed in.
    cyg_uint32         segMask  = (cyg_uint32)_regionend_ & 0xf0000000;

    if (memTop == NULL)
    {
        // We reserve 4k at the top of RAM, for no good reason.  VxWorks reserved
        // 64k, apparently for the vxWorks boot loader, which we aren't using.
        memTop = (cyg_uint8 *)(MIN( CalculateRamSize(), hal_mips_mem_limit ) - (4*1024) + segMask);
    }
    return memTop;
}


// The watch-dog is initialized by setting the counter value register
// However, the watch-dog control register needs to be read to get the
// current counter value. Reading the counter value register simply
// returns the original value that was written to it.

extern cyg_uint32 gWatchDogMinimumValue;
extern void     (*gWatchDogResetCallout)(void);

void hal_check_watchdog_timer( void );
void hal_check_watchdog_timer( void )
{
    static bool CalloutCalled = false;

    if (TIMER->WatchDogCtl < gWatchDogMinimumValue)
    {
        if (gWatchDogResetCallout != 0 && !CalloutCalled)
        {
            (*gWatchDogResetCallout)();
            CalloutCalled = true;
        }
    }
}


/*------------------------------------------------------------------------*/
/* End of plf_misc.c                                                      */
