//****************************************************************************
//
//  Copyright (c) 2003  Broadcom Corporation
//  All Rights Reserved
//  No portions of this material may be reproduced in any form without the
//  written permission of:
//          Broadcom Corporation
//          16215 Alton Parkway
//          Irvine, California 92618
//  All information contained in this document is Broadcom Corporation
//  company private, proprietary, and trade secret.
//
//****************************************************************************
//  $Id: if_bcm.h 1.16 2005/11/14 18:27:34 cliffd Exp $
//
//  Filename:       if_bcm.h
//  Author:         Cliff Danielson
//  Creation Date:  Aug 04, 2003
//
//****************************************************************************
//  Description:
//      Declarations for Broadcom network interface.
//
//****************************************************************************

//********************** Include Files ***************************************

// begin cliffmod
#ifndef __IF_BCM_H__
#define __IF_BCM_H__

#include <stdio.h>

    
#ifdef __cplusplus
extern "C" {
#endif

/* JRM */
#if 1 //0
#define stack_debug printf
#else
static void DummyPrint(const char *pString, ...) {}

#define stack_debug DummyPrint
#endif

struct ip_moptions;
struct route;
struct mbuf;
struct ifnet;
struct ifaddr;
struct sockaddr;
struct inpcb;
struct sockaddr_in;
struct ip;

/* DMP - the modules/functions that have debug that can be enabled/disabled */
#define DMP_DEBUG_IP_OUTPUT             0x0001
#define DMP_DEBUG_IP_OUTPUT_DIRECT      0x0002
#define DMP_DEBUG_RIP_OUTPUT            0x0004

/* DMP - bitmask of debugging to be enabled. */
#define DMP_STACK_DEBUG 0 //(DMP_DEBUG_IP_OUTPUT | DMP_DEBUG_IP_OUTPUT_DIRECT | DMP_DEBUG_RIP_OUTPUT)

// maximum of 8 network interfaces are supported at this time.
#define BCM_MAX_NET_INTERFACES 8 
#define BCM_MIN_NETIF_NUM 1
#define BCM_MAX_NETIF_NUM BCM_MAX_NET_INTERFACES

// application code identifies desired network interface by specifying
// one based values, i.e. network interface 1 thru 8.  a value of 0
// means that the application is not specifying a network interface.
//
// socket option bits are used to store the desired network interface.
// we must represent values nine values (default value of no preference, 
// and net if 1 thru 8) in unused socket option bits.  ideally we would
// use a continous value range to indicate broadcom network interface 
// number.  unfortunately most socket option bits are unavailable 
// for our use.  using conventions of eCos FreeBSD IP stack, 
// socket option bits 11, 13, 14 and 15 are available for our use.
// given that we wish to encode (9) network interface values,
// we need to use all four available bits.  Dave P. recommends
// changing convention such that interface 0 is the default
// interface and represented by value of 0 in the relevant
// option bits.
//                                              binary
#define BCM_NETIF_SOCKOPT_BITS  0xE800  //      1 1 1 x  1 x x x  x x x x  x x x x
#define BCM_NETIF_ANY           0x0000  //      0 0 0 x  0 x x x  x x x x  x x x x
#define BCM_NETIF_1             0x0800  //      0 0 0 x  1 x x x  x x x x  x x x x
#define BCM_NETIF_2             0x2000  //      0 0 1 x  0 x x x  x x x x  x x x x
#define BCM_NETIF_3             0x4000  //      0 1 0 x  0 x x x  x x x x  x x x x
#define BCM_NETIF_4             0x6000  //      0 1 1 x  0 x x x  x x x x  x x x x
#define BCM_NETIF_5             0x8000  //      1 0 0 x  0 x x x  x x x x  x x x x
#define BCM_NETIF_6             0xA000  //      1 0 1 x  0 x x x  x x x x  x x x x
#define BCM_NETIF_7             0xC000  //      1 1 0 x  0 x x x  x x x x  x x x x
#define BCM_NETIF_8             0xE000  //      1 1 1 x  0 x x x  x x x x  x x x x



struct pifnet_result
{
    // function specific result code.
    int result_code;

    // if function is successful, pointer to network interface result.
    struct ifnet* pifnet;

    // if function is successful, pointer to network interface address result.
    struct ifaddr* pifaddr;        
};

// BcmTestAndComputeLocalNetIf() - Test whether or not Broadcom 
//      specific IP output routing applies, and if it does,
//      then determine the outbound network interface
//      return pointers to associated ifnet structure and 
//      IP address.
//
// Params:
//      pinpcb - pointer to protocol control block associated with socket.
//      pifnet - pointer to ifnet structure that is to be
//          loaded with value for the specified network interface.
//      pif_ipaddr - pointer to IP address that is to be 
//          loaded with value for the specified network interface.
//
// Returns:
//      int - returns non-zero if a network interface was 
//          specified and computations were successful.
//
int BcmTestAndComputeLocalNetIf( register struct inpcb *pinpcb, 
    struct ifnet* pifnet, unsigned long* pif_ipaddr );


// BcmTestAndApplyUdpOutputRouting() - Test whether or not Broadcom specific 
//      UDP output routing applies, and if it does, then route the packet.
//
// Parameters:
//      pinpcb - pointer to protocol control block associated with socket.
//      pmbuf - pointer to data mbuf.
//      perror - pointer to error variable to load with result of routing
//          operation.
//
// Returns:
//      non-zero if packet has been routed by Broadcom specific mechanism.
//      
unsigned int BcmTestAndApplyUdpOutputRouting( register struct inpcb *pinpcb, 
    struct mbuf *pmbuf, int *perror );
    

// BcmTestAndApplyRawIpOutputRouting() - Test whether or not Broadcom specific 
//      raw IP output routing applies, and if it does, then route the packet.
//
// Parameters:
//      pinpcb - pointer to protocol control block associated with socket.
//      pmbuf - pointer to data mbuf.
//      perror - pointer to error variable to load with result of routing
//          operation.
//
// Returns:
//      non-zero if packet has been routed by Broadcom specific mechanism.
//      
unsigned int BcmTestAndApplyRawIpOutputRouting( register struct inpcb *pinpcb, 
    struct mbuf *pmbuf, int *perror );
                                 

// BcmAltIpOutputRouting() - Output IP packet according to Broadcom specific
//      routing rules.  This function is an alternative to the traditional
//      ip_output() function.
//
// Parameters:
//      m0 - pointer to data mbuf.
//      opt - pointer to options mbuf.
//      ro - pointer to traditional route structure.
//      flags - control flags.
//      imo - pointer to ip options structure.
//      src_addr - IP address of network interface on which the
//          packet is to be transmitted.
//      iifp - pointer to network interface structure for interface
//          on which the packet is to be transmitted.
//
// Returns:
//      error code as used by traditional ip_output() function.
//      
int BcmAltIpOutputRouting( struct mbuf *m0, struct mbuf *opt, struct route *ro, 
    int flags, struct ip_moptions *imo, unsigned long src_addr, 
    struct ifnet *iifp );


// BcmDefIpOutputRoutingMods() - Apply Broadcom specific routing modifications
//      to the default route.  This function is called from the 
//      traditional ip_output() function.
//
// Parameters:
//      pip_packet - pointer to IP packet structure for packet to output.
//      proute - pointer to default route.
//      pif_pair - if default route is modified, this param is loaded
//          with pointers to the ifnet and ifaddr for the outbound
//          network interface.
//
// Returns:
//      struct pifnet_result.
//      if route is not changed, then
//          pifnet_result.result_code = RTMODS_NO_CHANGE
//          pifnet_result.pifnet = points to ifnet for outbound net interace
//          pifnet_result.pifaddr = points to outbound net interface address
//      else if packet routed to different network interface, then
//          pifnet_result.result_code = RTMODS_DIFF_NET_IF
//          pifnet_result.pifnet = points to ifnet for outbound net interace
//          pifnet_result.pifaddr = points to outbound net interface address
//      else 
//          assume route could not be determined...
//          pifnet_result.result_code = RTMODS_FAILED
//          pifnet_result.pifnet = 0
//          pifnet_result.pifaddr = 0

#define RTMODS_NO_CHANGE 0
#define RTMODS_FAILED    1 
#define RTMODS_DIFF_NET_IF 2
//
struct pifnet_result BcmDefIpOutputRoutingMods( struct ip* pip_packet, struct route* proute );


// BcmComputeDefaultGatewayIpAndMacAddr() - Given pointers to outbound network
//      interface structure and packet, compute default gateway IP and MAC
//      addresses.
//
// Parameters:
//      pifnet - pointer to ifnet structure for outbound network interface.
//      ppacket_buf - pointer to outbound packet buffer.
//      pgateway_ip_addr - pointer to IP address of network interface.
//      pgateway_mac_addr - pointer to MAC address of network interface.
//
// Returns:
//      non-zero if computation is successful.
//
int BcmComputeDefaultGatewayIpAndMacAddr( register struct ifnet *pifnet,
	struct mbuf *ppacket_buf, struct sockaddr_in *pgateway_ip_addr, 
    char *pgateway_mac_addr );


// BcmNetIfNameToSockOpt() - Given network interface name, compute and return
//      Broadcom socket option bits associated with the specified net if name.
//
// Params:
//      net_if_name - network interface name.
//
// Returns:
//      unsigned short - Broadcom defined socket option value that is
//          associated with the interface.  If the network interface is 
//          unknown, the default value of 0 is returned.
//
unsigned short BcmNetIfNameToSockOpt( const char* net_if_name );


// BcmNetIfNumToSockOpt() - Given network interface number, compute and return
//      Broadcom socket option bits associated with the specified net if number.
//
// Params:
//      net_if_num - Broadcom network interface number.  valid range
//          is BCM_MIN_NETIF_NUM to BCM_MAX_NETIF_NUM.
//
// Returns:
//      unsigned short - Broadcom defined socket option value that is
//          associated with the interface.  If the network interface is 
//          unknown, the default value of 0 is returned.
//
unsigned short BcmNetIfNumToSockOpt( unsigned int net_if_num );


// BcmNetIfNameToNetIfNum() - Given network interface name, compute and return
//      Broadcom network interface number.
//
// Params:
//      net_if_name - network interface name.
//
// Returns:
//      int - return value in the range of 0 to BCM_MAX_NET_IF_NUM.   
//              undefined socket option bits mappings will be assigned
//              to default network interface 0.
//
unsigned int BcmNetIfNameToNetIfNum( const char* net_if_name );


// BcmNetIfNumIsValid() - Returns non-zero if specified Broadcom network
//      interface number is valid, and returns zero if it is invalid.
//
unsigned int BcmNetIfNumIsValid( unsigned int net_if_num );


// BcmNetIfSpecified() - Returns non-zero if socket options parameter
//      specifies a Broadcom specific network interface.
//
unsigned int BcmNetIfSpecified( unsigned short socket_options );


// BcmNetIfNumToNetIfPtr() - Given Broadcom network interface number, 
//      lookup and return pointer to associated network interface structure.
//
// Params:
//      net_if_num - Broadcom network interface number.  valid range
//          is BCM_MIN_NETIF_NUM to BCM_MAX_NETIF_NUM.
//
// Returns:
//      ifnet* - pointer to associated network interface structure.
//
struct ifnet* BcmNetIfNumToNetIfPtr( unsigned int net_if_num );


// BcmSetNetIfPtr() - Set the associated network interface structure 
//      pointer for the specified Broadcom network interface number.
//
// Params:
//      net_if_num - Broadcom network interface number.  valid range
//          is BCM_MIN_NETIF_NUM to BCM_MAX_NETIF_NUM.
//
//      pifnet - pointer to associated network interface structure.
//
void BcmSetNetIfPtr( unsigned int net_if_num, struct ifnet* pifnet );


// BcmSetNetIfIpAddr() - Set the 32 bit representation of the network 
//      interface IP address for the specified Broadcom network interface 
//      number.
//
// Params:
//      net_if_num - Broadcom network interface number.  valid range
//          is BCM_MIN_NETIF_NUM to BCM_MAX_NETIF_NUM.
//
//      net_if_ip_addr - 32 bit representation of the network interface 
//          IP address
//
void BcmSetNetIfIpAddr( unsigned int net_if_num, unsigned long net_if_ip_addr );


// BcmSetNetIfPtrAndIpAddr() - Initialize Broadcom network interface 
//      struct pointer, bind name to interface, and store IP address 
//      of network interface.
//
// Params:
//      net_if_num - Broadcom network interface number.  valid range
//          is BCM_MIN_NETIF_NUM to BCM_MAX_NETIF_NUM.
//
//      net_if_ip_addr - 32 bit representation of the network interface 
//          IP address
//
void BcmSetNetIfPtrAndIpAddr( unsigned int net_if_num, unsigned long net_if_ip_addr );


// BcmNetIfNumToName() - Given Broadcom network interface number, 
//      lookup and return the name of the network interface.
//
// Params:
//      net_if_num - Broadcom network interface number.  valid range
//          is BCM_MIN_NETIF_NUM to BCM_MAX_NETIF_NUM.
//
// Returns:
//      char* - character string representation of the network interface name.
//
char* BcmNetIfNumToName( unsigned int net_if_num );


// BcmSockOptToNetIfNum() - Given socket options, compute and return
//      network interface number.
//
// Params:
//      socket_options - 16 bit socket options.
//
// Returns:
//      int - return value in the range of 0 to BCM_MAX_NET_IF_NUM.   
//              undefined socket option bits mappings will be assigned
//              to default network interface 0.
//
unsigned int BcmSockOptToNetIfNum( unsigned short socket_options );


// BcmNetIfGetFirstIpAddrViaIoctl() - Given network interface name,
//      get the first IP address associated with the net if via system 
//      ioctl call.
//
// Parameters:
//      pnet_ifname - pointer to network interface name.
//
// Returns:
//      unsigned long - first IP address of network interface in 32 bit
//          format.
//
unsigned long BcmNetIfGetFirstIpAddrViaIoctl( char* pnet_if_name );


// Currently, we only support 8 IP stacks, and we want one gateway for each.
#define MAX_DEFAULT_GATEWAYS 8

// BcmSetDefaultIpGateway() - Store association of specified IP stack instance
//      and gateway router IP address.
//
// Parameters:
//      stackIndex - IP stack identifier (is this the same as Broadcom
//          network interface number???)
//
//      routerAddress - IP address of default gateway router for the 
//          specified IP stack.
//
void BcmSetDefaultIpGateway(unsigned int stackIndex, unsigned long routerAddress);


// BcmGetDefaultIpGateway() - Return gateway router IP address associated
//      with the specified IP stack instance.
//
// Parameters:
//      stackIndex - IP stack identifier (is this the same as Broadcom
//          network interface number???)
//
// Returns: 
//      unsigned long - IP address of default gateway router for the 
//          specified IP stack.
//
unsigned long BcmGetDefaultIpGateway(unsigned int stackIndex);


// BcmSetDefaultIpGatewayMacAddr() - Store association of specified IP stack 
//      instance and gateway router MAC address.
//
//      NOTE: This function must be called for network interfaces
//      with contrived gateways, such as the interface used for 
//      the built in web server and CPE side SNMP access.  
//      In other words, this function is called instead of
//      configuring a static ARP table entry in the IP stack
//      ARP/routing tables.
//
// Params:
//      stackIndex - IP stack identifier (is this the same as Broadcom
//          network interface number???)
//
//      gateway_if_mac_addr - char* buffer containing 48 bit representation 
//          of the default gateway mac address for the target interface.
//
void BcmSetDefaultIpGatewayMacAddr( unsigned int stackIndex, const char* gateway_mac_addr );


// BcmGetDefaultIpGatewayMacAddr() - If default gateway MAC address is specified
//      for the target stackIndex, then copy the MAC address value into
//      memory pointed to by gateway_mac_addr param.
//
// Params:
//      stackIndex - IP stack identifier (is this the same as Broadcom
//          network interface number???)
//
//      gateway_if_mac_addr - char* to target buffer for 48 bit representation 
//          of the default gateway mac address.
//
// Returns:
//      int - non-zero return indicates success.  zero return indicates
//      that a default gateway MAC address has NOT been sppecified
//      for this network interface.
//          
int BcmGetDefaultIpGatewayMacAddr( unsigned int stackIndex, char* gateway_mac_addr );


// BcmDeleteDefaultIpGatewayMacAddr() - Store association of specified IP stack 
//      instance and gateway router MAC address.
//
// Params:
//      stackIndex - IP stack identifier (is this the same as Broadcom
//          network interface number???)
//    
void BcmDeleteDefaultIpGatewayMacAddr( unsigned int stackIndex );


// BcmIsIpAddrOnThisLocalNetIf() - returns non-zero if the target IP address 
//      associated with the specified local network interface.
//
// Parameters:
//      target_ip_u32 - target IP address to look for on the network interface.
//      pifnet - pointer to network interface that is to be searched.
//
// Returns:
//      if search is successful, then pointer to the ifaddr structure 
//          which contains the matching local interface address
//          is returned.  otherwise 0 is returned.
//
struct ifaddr* BcmIsIpAddrOnThisLocalNetIf( unsigned long target_ip_u32, 
    struct ifnet *pifnet );


// BcmIsIpAddrOnAnyLocalNetIf() - returns non-zero if the target IP address is
//      associated with any local network interface.
//
// Parameters:
//      target_ip_u32 - target IP address to match.
//      pif_pair - if search is successful, this param
//          set to point to ifnet and ifaddr for the matching
//          network interface.
//
// Returns:
//      pifnet_result structure.
//      if a matching local network interface is found, then
//          pifnet_result.result_code is non-zero
//          pifnet_result.pifnet points to matching network interface
//              ifnet structure
//          pifnet_result.pifaddr points to matching network interface
//              address ifaddr structure.
//      else
//          pifnet_result.result_code is 0
//          pifnet_result.pifnet is 0
//          pifnet_result.pifaddr is 0
//
struct pifnet_result BcmIsIpAddrOnAnyLocalNetIf( unsigned long target_ip_u32 ); 
    
    
// BcmShowDefaultIpGateway() - Print associations of IP stack instance id
//      and gateway router IP address.
//
void BcmShowDefaultIpGateway(void);


// BcmShowNetworkTables() - Calls eCos supplied show_network_tables() 
//      function with simpler syntax.
//
void BcmShowNetworkTables(void);


// BcmShowRoutingTable() - Customizes eCos supplied show_network_tables() 
//      function such that only routing table info is output.
//
void BcmShowRoutingTable(void);


// BcmShowArpTable() - Customizes eCos supplied show_network_tables() 
//      function such that only ARP table info is output.
//
void BcmShowArpTable(void);


// BcmShowNetworkMemStats() - Calls eCos support function 
//      cyg_kmem_print_stats(), which prints info regarding 
//      network stack mbuf usage.
//
void BcmShowNetworkMemStats(void);


// BcmShowMbuf() - Print the specified mbuf chain to the console.
//
void BcmShowMbuf( struct mbuf *pmbuf );


// BcmFindArpEntry() - Search the ARP table for an entry with the
//      specified IP address.  If a match is found, load the associated
//      MAC address into the targetMacAddr passed param.
//
// Parameters:
//      targetIpAddress - known target IP address
//
//      targetMacAddr - unknown target MAC address
//
// Returns:
//      non-zero - arp entry found
//      false - no arp entry found 
//
// Note:
//     Some of this function was taken from rn_walktree
//
unsigned int BcmFindArpEntry( unsigned int targetIpAddress, char *targetMacAddr );


// BcmAddArpIpToMacAssoc() - Add specified IP to MAC address association
//      to the IP stack ARP table. 
//
// Parameters:
//      ipAddress - 32 bit IP address
//
//      macAddress - associated 48 bit ethernet MAC address
//
// Returns:
//      int - value of 0 indicates success.
//
int BcmAddArpIpToMacAssoc( unsigned long ipAddress, unsigned char* macAddress );


// BcmDeleteArpIpToMacAssoc() - Delete specified IP to MAC address association
//      to the IP stack ARP table. 
//
// Parameters:
//      ipAddress - 32 bit IP address
//
//      macAddress - associated 48 bit ethernet MAC address
//
// Returns:
//      int - value of 0 indicates success.
//
int BcmDeleteArpIpToMacAssoc( unsigned long ipAddress, unsigned char* macAddress );


// ethmac_ntoa() - Helper function to convert ethernet MAC address 
//      from network to ASCII presentation format.
//
char* ethmac_ntoa( char* ethmac_addr );


// Functions to manipulate bit mapped options for TCP/IP stack debug logging.
//
// Note: eCos defined bit values for the low order (16) bits are defined in
//      \packages\net\bsd_tcpip\v2_0\include\sys\param.h
//
// Broadcom defined values for higher order bit values are defined in this file.
//
// BcmGetNetLogMask() - returns the current value for cyg_net_log_mask.
//
// BcmNetLogMaskSet() - assigns the specified value to cyg_net_log_mask.
//      returns the prior value of cyg_net_log_mask.
//
// BcmNetLogMaskSetBits() - logical ors the specified value with the current
//      value of cyg_net_log_mask.  returns the new value of cyg_net_log_mask.
//
// BcmNetLogMaskClearBits() - logical xors the specified value with the current
//      value of cyg_net_log_mask.  returns the new value of cyg_net_log_mask.
//
unsigned int BcmNetLogMaskGet();
unsigned int BcmNetLogMaskSet( unsigned int newValue );
unsigned int BcmNetLogMaskSetBits( unsigned int bitMask );
unsigned int BcmNetLogMaskClearBits( unsigned int bitMask );
//
// Broadcom specific logging options.
//
// log buffers for receive data path at entry to functions:
#define BCM_LOG_MBUF_SORCV              0x00010000  // soreceive() 
#define BCM_LOG_MBUF_TCP_INPUT          0x00020000  // tcp_input()
#define BCM_LOG_MBUF_IP_INPUT           0x00080000  // ip_input()
//
// routing table internal details included in "show network tables" output
#define BCM_LOG_ROUTING_TABLE_DETAILS   0x00100000
//
// 
// file.


#ifdef __cplusplus
}
#endif


#endif /* __IF_BCM_H__ */

// end cliffmod

